/*
 * Author: Dušan Kolář
 */

#include "parse_mms.h"

#include <stddef.h>

#define MIN_LENGTH_ALL      26

#define COTP_TYPE_MAGIC   0xf0
#define SPDU_TYPE_MAGIC   0x01
#define PPDU_TYPE_MAGIC   0x61
#define PDV_LIST_MAGIC    0x30
#define ASN_TYPE_MAGIC    0xa0

#define MMS_TYPE_MIN      0xa0
#define MMS_TYPE_MAX      0xad

#define IS_LONG_DATA(x)   ((x)==0xa1 || (x)==0xa0)
#define IS_SHORT_DATA(x)  ((x)==0xa3)


#define SWAP_TYPE(WIDTH) typedef union u_swap ## WIDTH {\
    uint ## WIDTH ## _t value;\
    uint8_t  bytes[WIDTH/8];\
} swap ## WIDTH ## _t;

SWAP_TYPE(64)
SWAP_TYPE(32)
SWAP_TYPE(16)


// .................................................................
// local function declaration

static mms_parse_error_t tlv_len(uint8_t **input, uint16_t *len, uint32_t *size);

// .................................................................
// local data

static mms_data_t data_res;

// .................................................................
// ------------------------ GLOBAL FUNCTIONS ------------------------


mms_data_t *parse_mms(uint8_t *input, uint32_t size, mms_parse_error_t *error) {
    // total length is too short to put in minimal information for GOOSE
    if (size < MIN_LENGTH_ALL) {
        *error = MMSPE_TOO_SHORT;
        return (NULL);
    }
    
    // reset result structure
    data_res.mms_type_tag = 0;
    data_res.service_flag = 0;
    data_res.service_type_tag = 0;
    
    input += 2; // skip TPKT version and reserved byte
    
    // read and test TPKT length
    swap16_t tpkt_len;
    tpkt_len.bytes[1] = *input++;
    tpkt_len.bytes[0] = *input++;
    
    if ((uint32_t)tpkt_len.value != size) {
        // lengths don't match
        *error = MMSPE_LENGTH_MISMATCH;
        return (NULL);
    }
    size -= 4;
    
    // skip to COTP type magic
    ++input;
    --size;
    
    if (*input != COTP_TYPE_MAGIC) {
        *error =  MMSPE_COTP_MAGIC_FAIL;
        return (NULL);
    }

    // skip to SPDU type magic
    input += 2;
    size -= 2;
    
    if (*input != SPDU_TYPE_MAGIC) {
        *error =  MMSPE_SPDU_MAGIC_FAIL;
        return (NULL);
    }

    // skip to PPDU type magic
    input += 4;
    size -= 4;

    if (*input++ != PPDU_TYPE_MAGIC) {
        *error =  MMSPE_PPDU_MAGIC_FAIL;
        return (NULL);
    }
    --size;

    // PPDU len read    
    uint16_t tmp_len;
    mms_parse_error_t err;
    
    if ((err = tlv_len(&input, &tmp_len, &size)) != MMSPE_OK) {
        *error = err;
        return (NULL);
    }
    
    if (size != tmp_len) {
        // lengths don't match
        *error = MMSPE_LENGTH_MISMATCH;
        return (NULL);
    }

    // PDV list test
    if (*input++ != PDV_LIST_MAGIC) {
        *error =  MMSPE_PDV_LIST_MAGIC_FAIL;
        return (NULL);
    }
    --size;

    
    // PDV list length
    if ((err = tlv_len(&input, &tmp_len, &size)) != MMSPE_OK) {
        *error = err;
        return (NULL);
    }

    if (size != tmp_len) {
        // lengths don't match
        *error = MMSPE_LENGTH_MISMATCH;
        return (NULL);
    }

    if (size < 5) {
        // packet size too low
        *error = MMSPE_LENGTH_MISMATCH;
        return (NULL);
    }

    // skip to single ASN1 type taga
    input += 3;
    size -= 3;
    
    // PDV list test
    if (*input++ != ASN_TYPE_MAGIC) {
        *error =  MMSPE_ASN_MAGIC_FAIL;
        return (NULL);
    }
    --size;

    
    // MMS length
    if ((err = tlv_len(&input, &tmp_len, &size)) != MMSPE_OK) {
        *error = err;
        return (NULL);
    }
    
    if (size != tmp_len) {
        // lengths don't match
        *error = MMSPE_LENGTH_MISMATCH;
        return (NULL);
    }

    if (size < 3) {
        // packet size too low
        *error = MMSPE_LENGTH_MISMATCH;
        return (NULL);
    }
    
    
    // MMS tag test
    if (*input<MMS_TYPE_MIN || *input>MMS_TYPE_MAX) {
        *error =  MMSPE_MMS_MAGIC_FAIL;
        return (NULL);
    }
    data_res.mms_type_tag = *input++;
    --size;
    
    
    // LONG MMS data structure
    if (IS_LONG_DATA(data_res.mms_type_tag)) {
        
        if (size < 7) {
            // packet size too low
            *error = MMSPE_LENGTH_MISMATCH;
            return (NULL);
        }
        
        // payload length
        if ((err = tlv_len(&input, &tmp_len, &size)) != MMSPE_OK) {
            *error = err;
            return (NULL);
        }

        if (size != tmp_len) {
            // lengths don't match
            *error = MMSPE_LENGTH_MISMATCH;
            return (NULL);
        }
        
        if (size < 6) {
            // packet size too low
            *error = MMSPE_LENGTH_MISMATCH;
            return (NULL);
        }

        // skip to service type
        input += 4;
        size -= 4;
        
        data_res.service_flag = 1;
        data_res.service_type_tag = *input;
                
    } else
    
        
    // SHORT MMS data structure
    if (IS_SHORT_DATA(data_res.mms_type_tag)) {
        
        if (size < 3) {
            // packet size too low
            *error = MMSPE_LENGTH_MISMATCH;
            return (NULL);
        }
        
        // payload length
        if ((err = tlv_len(&input, &tmp_len, &size)) != MMSPE_OK) {
            *error = err;
            return (NULL);
        }

        if (size != tmp_len) {
            // lengths don't match
            *error = MMSPE_LENGTH_MISMATCH;
            return (NULL);
        }
        
        if (size < 2) {
            // packet size too low
            *error = MMSPE_LENGTH_MISMATCH;
            return (NULL);
        }

        data_res.service_flag = 1;
        data_res.service_type_tag = *input;        
    }
    
    *error = MMSPE_OK;
    return (&data_res);
}


// .................................................................
// ------------------------ LOCAL FUNCTIONS ------------------------


static mms_parse_error_t tlv_len(uint8_t **input, uint16_t *len, uint32_t *size) {
    uint8_t *in = *input;
    uint8_t enc = *in++;

    if (*size < 1) {
        *len   = 0;
        return (MMSPE_OUT_OF_DATA);
    }
    
    if (enc<128) { // short len, just this one byte carries length information
        *input = in;
        *len   = enc;
        --(*size);  // just one encoding byte read
        return (MMSPE_OK);
    }
    enc = enc & 0x7f;
    if (enc == 0) { // indefinite length, failure
        *input = in;
        *len   = 0;
        return (MMSPE_TLV_INDEFINITE);
    }
    if (enc > 2) { // number of length bytes > 2, probably error
        *input = in;
        *len = 0;
        return (MMSPE_TLV_LARGE);
    }
    
    if (*size < 2) {
        *input = in;
        *len   = 0;
        return (MMSPE_OUT_OF_DATA);
    }
    
    if (enc == 1) {
        *len = (uint16_t)(*in++);
        *input = in;
        (*size) -= 2;  // encoding and 1 length byte
        return (MMSPE_OK);
    }
    
    if (*size < 3) {
        *input = in;
        *len   = 0;
        return (MMSPE_OUT_OF_DATA);
    }
    
    // remains length 2, should change BE -> LE
    swap16_t sw_len;
    sw_len.bytes[1] = *in++;
    sw_len.bytes[0] = *in++;
    *len = sw_len.value;
    *input = in;
    (*size) -= 3;  // encoding and 2 length bytes
    return (MMSPE_OK);
}


// EOF parse_goose
// .................................................................
// .................................................................
